#######################################################################
#
# state component
#
# NCM access control
#
# Miroslav Siket <Miroslav.Siket@cern.ch>
# Version 1.0.0 : 01/06/04 12:18
#
# Copyright (c) 2004 Miroslav Siket, CERN and EU DataGrid.
# For license conditions see http://www.eu-datagrid.org/license.html
#
#######################################################################

package NCM::Component::access_control;

#
# a few standard statements, mandatory for all components
#

use strict;
use NCM::Component;
use vars qw(@ISA $EC);
@ISA = qw(NCM::Component);
$EC=LC::Exception::Context->new->will_store_all;

use LC::File qw(copy);
use NCM::Check;

##########################################################################
sub Compare_arrays {
##########################################################################

	my ($self, $array1, $array2) = @_;
	my $ret_value = "true";
	
	my @arr1 = @{$array1};
	my @arr2 = @{$array2};
	my @empty = ();

        my @union = ();
	my @intersection = ();
	my @difference = ();
        my %count = ();

	my $element;
        foreach $element (@arr1, @arr2) { $count{$element}++ }

        foreach $element (keys %count) {
		push @union, $element;
		push @{ $count{$element} > 1 ? \@intersection : \@difference }, $element;
    	}
	
	if(@difference eq @empty){
		return "true";
	} else {
		$self->log("difference: @difference");
		return "false";
	}

} 

##########################################################################
sub Get_acl {
##########################################################################
	my $config;
	my $acl_link;
	($config, $acl_link) = @_;
	my @return_array = ();
	my $it;
	my $it_n;

	unless ($config->elementExists($acl_link)) {
		return @return_array;
	} else {
		$it = $config->getElement($acl_link);
		while($it->hasNextElement()){
			$it_n = $it->getNextElement();
			push(@return_array,$it_n->getValue());	
		}
	}
	@return_array = sort(@return_array);
	return @return_array;
}

##########################################################################
sub Get_acl_file {
##########################################################################
	my $file_r;
	my $sep_string = "";
	my $file_h;
	my @empty = ();
	my @item;
	my @found_elem;

	($file_r, $sep_string) = @_;
	my @return_array = ();
	if (!open(file_h, $file_r)){
		return @return_array;
	}
	while (<file_h>) {
		@item = split(/$sep_string/);
		#@found_elem = grep(/$item[0]/i, @return_array);
		#if(@found_elem eq @empty ){
		push(@return_array,$item[0]);
		#}
	}
	close file_h;
	@return_array = sort(@return_array);
	return @return_array;
}

##########################################################################
sub session_file {
##########################################################################

	my ($acl_f, $tmp_acl) = @_;
	my $return_v = "";
	my @acl = @{$tmp_acl};

	my %hrights;
	my %maxlines;
	my $oid;
	my $nid;
	my $line;
	my $tcontent = "";
	
	map {$hrights{$_} = "$_\t\t-\tmaxlogins\t9999\n"} @acl;
	
	# read in the file, if exists
	if (open(LIMITS, $acl_f)){
		my $first_maxlog_line = 0;
		while (<LIMITS>){
			$line = $_;
			if (($oid) = /^([^\s]+)\s+-\s+maxlogins.+\n/){
				if ($first_maxlog_line eq 0){
					$return_v .= $line;
					$first_maxlog_line = 1;
				}
				if ($hrights{$oid}) {
					$maxlines{$oid} = $hrights{$oid};
					delete($hrights{$oid});
				}
			} else {
				$return_v .= $line;	
			}
		}
		close LIMITS;
		
		# create content of the new file
		foreach $nid (keys(%hrights)){
			if($hrights{$nid}){
				$maxlines{$nid} = $hrights{$nid};
			}
		}
		map {$tcontent .= "$maxlines{$_}"} sort keys(%maxlines);
		$tcontent .= "*\t\t-\tmaxlogins\t0\n";
		if($first_maxlog_line eq 1){
			$return_v=~ s{^.+maxlogins.+\n}{$tcontent}mg;
		} else {
			$return_v .= $tcontent;
		}
	} else {
		map {$return_v .= "$hrights{$_}"} sort keys(%hrights);
		$return_v .= "*\t\t-\tmaxlogins\t0\n";	
	} # - end of open file

	return $return_v;
}

##########################################################################
sub get_keys {
##########################################################################
	my ($self, $url, $ssh_dir, $acl_d) = @_;
	my @acl_r = @{$acl_d};
	my $return_v = "";
	my $r;

	foreach $r (@acl_r){
		my $keycontent = undef;
		if (eval "require LWP::Simple"){
			my $url_d = "$url/$r.key";
			$keycontent = LWP::Simple::get($url_d);			
		}
		if (not $keycontent){
			open(KEYFILE, "<$ssh_dir/$r.key") || next;
			read KEYFILE, $keycontent, 1024;
			close KEYFILE;
		}

		if (not $keycontent){
			$self->log("Unable to get ssh key for user: $r");
		}
		$return_v .= $keycontent if defined $keycontent;
	}
	return $return_v;
}


##########################################################################
sub Configure {
##########################################################################
  my ($self,$config)=@_;

  # iterators
  my $it;
  my $it_n;

  my @empty = ();

  # some constants
  my $url = "http://ccconf.cern.ch/ssh-public-keys/";
  my $ssh_dir = "/afs/cern.ch/project/core/conf/ssh";
  my %acl_file;
  $acl_file{"login"} = "/etc/loginusers";
  $acl_file{"session"} = "/etc/security/limits.conf";
  my $roothome = (getpwnam("root"))[7];
  $acl_file{"keys"} = "$roothome/.ssh/authorized_keys";
  $acl_file{"klogin"} = "$roothome/.klogin";

  # store new content files into the structure
  my %content;
	
  my $contents;

  # try to find out if active
  my $statepath='/software/components/access_control/active';
  unless ($config->elementExists($statepath)) {
    $self->error("cannot get $statepath");
    return;
  }
  my $statestr=$config->getValue($statepath);
    
  if($statestr ne "true") {
	$self->warn("Access control is turned off.");
	return;
  };
	
  # get the interative field
  my $interactive = '/software/components/access_control/interactive_restrictions';
  my $interactive_rest = "";
  unless ($config->elementExists($interactive)) {
    $self->warn("interactive restrictions not defined, assuming active");
    $interactive_rest = "on";
  } else {
    $interactive_rest = $config->getValue($interactive);
  }

  # get kerberos field to fill in in case of .klogin entries
  my $krb_domainname_cdb = "/software/components/access_control/krb_domainname";
  my $krb_domainname = "";
  if ($config->elementExists($krb_domainname_cdb)){
	$krb_domainname = $config->getValue($krb_domainname_cdb);
  } else {
    $self->warn("Unable to determine kerberos domainname. Will not produce .klogin file");
  }

  # get the location of the ssh public keys 
  my $ssh_keys_location_cdb = "/software/components/access_control/ssh_keys_location";
  if($config->elementExists($ssh_keys_location_cdb)){
 	$url = 	$config->getValue($ssh_keys_location_cdb);
  } else {
    $self->warn("No locationof ssh public keys specified. Will not download the keys.");
  }
  

  # get list of users that do root access, interactive, console, sudo
  my $root_access = '/software/components/access_control/acl_root';
  my $inter_access = '/software/components/access_control/acl_interactive';
  my $console_access = '/software/components/access_control/acl_console';
  my $sudo_priv = '/software/components/access_control/acl_sudo';
  my $root_access_method = '/software/components/access_control/acl_root_method';
  my $inter_access_method = '/software/components/access_control/acl_int_method';

  # get root acl
  my @root_acl;
  @root_acl = Get_acl($config, $root_access);
  if (@root_acl eq @empty){
    $self->error("Root access list not defined.");
    return;
  }

  # get interactive acl
  my @interactive_acl;
  @interactive_acl = Get_acl($config, $inter_access);
  if (@interactive_acl eq @empty) {
    $self->warn("Interactive acl not defined, assuming none");
  }

  # get console acl
  my @console_acl;
  @console_acl = Get_acl($config, $console_access);
  if (@console_acl eq @empty) {
    $self->info("Console acl not defined, assuming none");
  } 

  # get sudo priv
  my %sudo_acl;
  unless ($config->elementExists($sudo_priv)) {
    $self->info("Sudo acl not defined, assuming none");
    %sudo_acl = ();
  } else {
  #    %sudo_acl = $config->getValue($sudo_priv);
    %sudo_acl = ();
  }

  # get type of root access type
  my $root_acl_method = "none";
  unless ($config->elementExists($root_access_method)) {
    $self->error("Root access method not defined.");
    return;
  } else {
     $root_acl_method = $config->getValue($root_access_method);
  }

  # get type of interactive access
  my $inter_acl_method = "none";
  unless ($config->elementExists($inter_access_method)) {
    $self->warn("Interactive access method not defined, assuming none");
  } else {
     $inter_acl_method = $config->getValue($inter_access_method);
  }

  # just get status in the log file
  $self->log("Component active: $statestr");
  $self->log("Interactive restrictions: $interactive_rest");
  $self->log("Root access methods: $root_acl_method");
  $self->log("Interactive access methods: $inter_acl_method");
  $self->log("Root acl: @root_acl");
  $self->log("Interactive acl: @interactive_acl");
  $self->log("Console acl: @console_acl");
  $self->log("Download directory of ssh public keys: $url");
  $self->log("Kerberos domainname: $krb_domainname");

  # read in the .klogin file and compare it with downloaded info
  my @klog_in = ();
  my $login;
  my $dom;
  my @found_elem;

  @klog_in = Get_acl_file($acl_file{"klogin"},'\.');
  if(@klog_in eq @empty){
	$self->warn("Cannot open $acl_file{'klogin'} or empty");
  }
  $self->log("Current klogin users: @klog_in");

  # compare the two
  my $not_do_acl_root = Compare_arrays($self,\@klog_in,\@root_acl);
  if ($not_do_acl_root eq "true"){
	$self->log("The acl from CDB is the same as current klogin acl for root.");	
  } else {
	$self->log("The acl from CDB is different from current klogin acl for root.");
  }
  
  # read in the /etc/loginusers file and compare it with downloaded info
  my @user_login_in = ();
  if ($interactive_rest eq "on"){
	@user_login_in = Get_acl_file($acl_file{"login"},'\n');

	if (@user_login_in eq @empty){
		$self->warn("Cannot open $acl_file{'login'} or empty");
	}

	$self->log("Current login users: @user_login_in");

	# compare the two
  	my $not_do_int_rest = Compare_arrays($self,\@user_login_in,\@interactive_acl);
  	if ($not_do_int_rest eq "true"){
		$self->log("The acl from CDB is the same as current interactive acl.");	
	  } else {
		$self->log("The acl from CDB is different from the current interactive acl.");
  	}
  }

  #######
  #
  # Time to check what needs to be done changes
  #
  #######

  # klogin
  if ($root_acl_method eq "all" or $root_acl_method eq "klogin"){
	if ($not_do_acl_root ne "true"){
  		if ($krb_domainname ne ""){
			map {$content{klogin} .= "$_".$krb_domainname."\n"} @root_acl;
		}
	} else {
		$self->log("No need to do changes to klogin. Seems current.");
	}
  }
  # $self->log("New content of .klogin file: $content{klogin}");

  # .ssh/keys
  if ($root_acl_method eq "all" or $root_acl_method eq "keys"){
	$content{keys} = get_keys($self, $url, $ssh_dir, \@root_acl);
  }
  # $self->log("New content of .keys file: $content{keys}");


  # session
  # modify the limits.conf file for any possible session access changes
  $content{session} = "";
  if ($interactive_rest eq "on") {
	if ($inter_acl_method eq "all" or $inter_acl_method eq "session"){
		$content{session} = session_file($acl_file{"session"},\@interactive_acl);
	}
  } else {
	$inter_acl_method = "not using";
  }
  # $self->log("New session file would be: $content{session}");

  # user login
  map {$content{login} .= "$_\n"} @interactive_acl;

  #######
  #
  # Writing the files
  #
  #######
  
  my $fi;
  my $lname;
  foreach $fi (keys(%content)){
	$lname = $acl_file{$fi};
	if ($content{$fi}){
		if(open(OUTFILE, ">$lname.new")){
			print OUTFILE $content{$fi};
			close OUTFILE;
		} else {
			$self->error("Unable to create file $lname.new");
			return;
		}
		# rename existing file to backup
		my $ret_c;
  		if (-w $lname){ 
			$ret_c = copy("$lname","$lname.old") 
		} else {
			$self->log("$lname file did not exist or couldn't move");
		}
		 
		# rename existing file to backup
		$ret_c = rename("$lname.new","$lname");
		if ($ret_c ne 1){
			$self->log("Unable to rename $lname.new to $lname");
		}
		unlink("$lname.new");
		$self->log("Modified $lname file.");
	}
  }

    # Create hooks for the access files in the PAM system
    # (for RH6 have to modify individual files, whilst RH7 uses PAM_stack)
    my @pamfiles;
    my $type;
    my $opath;
    my ($sysname, $sysver) = (POSIX::uname)[0,2];
    my @tmp = split(/\./,$sysver);
    $sysver = "$tmp[0]"."\."."$tmp[1]";
    $self->log("System name: $sysname and version: $sysver");
    if ($sysname eq "Linux" and $sysver lt 2.4) {
        @pamfiles = ("login", "rlogin", "rexec", "rsh", "xdm");
    } else {
        @pamfiles = ("system-auth");
    }
    if ($inter_acl_method eq "login" or $inter_acl_method eq "all") {
        $lname = $acl_file{"login"};
        foreach $type (@pamfiles) {
            NCM::Check::lines("/etc/pam.d/$type",
		"backup"  => ".old",
                "linere"  => ".*auth.*pam_listfile.*",
                "goodre"  => "auth       requisite    /lib/security/pam_listfile.so onerr=fail item=user sense=allow file=$lname",
                "good"    => "auth       requisite    /lib/security/pam_listfile.so onerr=fail item=user sense=allow file=$lname",
                "keep"    => "first",
                "add"    => "first"
            );
        }
    } else {
        unlink('$acl_file{"login"}');
        foreach $type (@pamfiles) {
            $opath = "/etc/pam.d/$type";
            LC::Check::file($opath,
		"source" => $opath,
                "code" => sub {
                 my($contents) = @_;
                 return() unless $contents;
                 $contents =~ s{^.+/lib/security/pam_listfile\.so.+\n}{}mg;
                 return($contents);
               }
            );
        }
    }
    if ($inter_acl_method eq "session" or $inter_acl_method eq "all") {
        $lname = $acl_file{"session"};
        foreach $type (@pamfiles) {
            NCM::Check::lines("/etc/pam.d/$type",
		"backup"  => ".old",
                "linere"  => ".*session.*pam_limits.*",
                "goodre"  => "session    requisite    /lib/security/pam_limits.so conf=$lname",
                "good"    => "session    requisite    /lib/security/pam_limits.so conf=$lname",
                "keep"    => "first",
                "add"    => "first"
            );
        }
    } else {
        $lname = $acl_file{"session"};
        # Delete lines from file
        LC::Check::file($lname,
	    "source" =>$lname,
            "code" => sub {
                my($contents) = @_;
                $contents =~ s{^.+maxlogins.+\n}{}mg;
                return($contents);
            }
        );
        foreach $type (@pamfiles) {
            $opath = "/etc/pam.d/$type";
            LC::Check::file($opath,
		"source" => $opath,
                "code" => sub {
                  my($contents) = @_;
                  return() unless $contents;
                  $contents =~ s{^.+/lib/security/pam_limits\.so.+\n}{}mg;
                  return($contents);
                }
            );
        }
    }

  return;
}

##########################################################################
sub Unconfigure {
##########################################################################
  my ($self,$config)=@_;
  return;
}

1; #required for Perl modules
